/**
 * Version 1.0 is to handle single variable 2DSquareCelled raster data.
 * Copyright (C) 2005 Andy Turner, CCG, University of Leeds, UK.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 */
package uk.ac.leeds.ccg.andyt.grids.exchange;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.math.BigDecimal;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellAbstract;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellDouble;
import uk.ac.leeds.ccg.andyt.grids.core.Grid2DSquareCellInt;
import uk.ac.leeds.ccg.andyt.grids.utilities.FileCreator;

/**
 * Class for exporting ESRI Asciigrid.
 */

//public class ESRIAsciiGridExporter extends ErrorHandler {
public class ESRIAsciiGridExporter {
    
    private static final long serialVersionUID = 1L;
    
    /** Creates a new instance of ESRIAsciiGridExporter */
    public ESRIAsciiGridExporter() {
        //this.initMemoryReserve();
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * a the File to which it was written.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell,
            boolean handleOutOfMemoryError ) {
        try {
            return toAsciiFile( grid2DSquareCell );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                grid2DSquareCell.clearMemoryReserve();
                grid2DSquareCell.swapToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
                grid2DSquareCell.initMemoryReserve();
                return toAsciiFile(
                        grid2DSquareCell,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * a the File to which it was written.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     */
    protected File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell ) {
        boolean handleOutOfMemoryError = false;
        File directory = grid2DSquareCell.getDirectory( handleOutOfMemoryError );
        File file = new File(
                directory.getParentFile(),
                grid2DSquareCell.getName( handleOutOfMemoryError ) + ".asc" );
        return toAsciiFile(
                grid2DSquareCell,
                file );
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell,
            File file,
            boolean handleOutOfMemoryError ) {
        try {
            return toAsciiFile(
                    grid2DSquareCell,
                    file );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                grid2DSquareCell.clearMemoryReserve();
                grid2DSquareCell.swapToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
                grid2DSquareCell.initMemoryReserve();
                return toAsciiFile(
                        grid2DSquareCell,
                        file,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     */
    protected File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell,
            File file ) {
        boolean handleOutOfMemoryError = false;
        return toAsciiFile(
                grid2DSquareCell,
                file,
                grid2DSquareCell.getNoDataValueBigDecimal( handleOutOfMemoryError ) );
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     * @param noDataValueBigDecimal The value to be used or substituted as a
     *   noDataValue for grid2DSquareCell.
     * @param handleOutOfMemoryError
     *   If true then OutOfMemoryErrors are caught, swap operations are initiated,
     *     then the method is re-called.
     *   If false then OutOfMemoryErrors are caught and thrown.
     */
    public File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell,
            File file,
            BigDecimal noDataValueBigDecimal,
            boolean handleOutOfMemoryError ) {
        try {
            return toAsciiFile(
                    grid2DSquareCell,
                    file,
                    noDataValueBigDecimal );
        } catch ( OutOfMemoryError oome0 ) {
            if ( handleOutOfMemoryError ) {
                grid2DSquareCell.clearMemoryReserve();
                grid2DSquareCell.swapToFileGrid2DSquareCellChunk( handleOutOfMemoryError );
                grid2DSquareCell.initMemoryReserve();
                return toAsciiFile(
                        grid2DSquareCell,
                        file,
                        noDataValueBigDecimal,
                        handleOutOfMemoryError );
            } else {
                throw oome0;
            }
        }
    }
    
    /**
     * Writes grid2DSquareCell out to file in ESRI Asciigrid format and returns
     * file.
     * @param grid2DSquareCell The Grid2DSquareCellAbstract for export.
     * @param file The File to export to.
     * @param noDataValueBigDecimal The value to be used or substituted as a
     *   noDataValue for grid2DSquareCell.
     */
    protected File toAsciiFile(
            Grid2DSquareCellAbstract grid2DSquareCell,
            File file,
            BigDecimal noDataValueBigDecimal ) {
        boolean handleOutOfMemoryError = true;
        String noDataValueBigDecimalString;
        if ( noDataValueBigDecimal == null ) {
            noDataValueBigDecimalString = "-9999.0";
        } else {
            noDataValueBigDecimalString = noDataValueBigDecimal.toString();
        }
        try {
            PrintWriter pw = new PrintWriter(
                    new BufferedWriter(
                    new OutputStreamWriter(
                    new FileOutputStream( file ) ) ) );
            BigDecimal[] dimensions = grid2DSquareCell.getDimensions( handleOutOfMemoryError );
            long nrows = grid2DSquareCell.getNrows( handleOutOfMemoryError );
            long ncols = grid2DSquareCell.getNcols( handleOutOfMemoryError );
            pw.println( "ncols " + ncols );
            pw.println( "nrows " + nrows );
            pw.println( "xllcorner " + dimensions[1].toString() );
            pw.println( "yllcorner " + dimensions[2].toString() );
            pw.println( "cellsize " + dimensions[0].toString() );
            
            if ( grid2DSquareCell.getClass() == Grid2DSquareCellInt.class ) {
                Grid2DSquareCellInt grid2DSquareCellInt = ( Grid2DSquareCellInt ) grid2DSquareCell;
                int noDataValue = grid2DSquareCellInt.getNoDataValue( handleOutOfMemoryError );
                pw.println( "NODATA_Value " + noDataValueBigDecimalString );
                long row;
                long col;
                int value;
                for ( row = nrows - 1; row > -1; row -- ) {
                    for ( col = 0 ; col < ncols; col ++ ) {
                        value = grid2DSquareCellInt.getCell(
                                row,
                                col,
                                handleOutOfMemoryError );
                        if ( value == noDataValue ) {
                            pw.print( noDataValueBigDecimalString + " " );
                        } else {
                            pw.print( value + " " );
                        }
                    }
                    pw.println( "" );
                }
            } else {
                //grid2DSquareCell.getClass() == Grid2DSquareCellDouble.class
                Grid2DSquareCellDouble grid2DSquareCellDouble = ( Grid2DSquareCellDouble ) grid2DSquareCell;
                double noDataValue = grid2DSquareCellDouble.getNoDataValue( handleOutOfMemoryError );
                if (  Double.isInfinite( noDataValue ) ) {
                    System.out.println(
                            "Warning!!! noDataValue = "
                            + noDataValue +
                            " in ESRIAsciigridExporter.toAsciiFile( \n" +
                            "Grid2DSquareCellAbstract( " + grid2DSquareCellDouble.toString( handleOutOfMemoryError ) + " ), \n" +
                            "File( " + file.toString() + " ) )" );
                }
                pw.println( "NODATA_Value " + noDataValueBigDecimalString );
                long row;
                long col;
                double value;
                for ( row = nrows - 1; row > -1; row -- ) {
                    for ( col = 0 ; col < ncols; col ++ ) {
                        try {
                            //pw.print( grid.getCell( row, col ) + " " );
                            value = grid2DSquareCellDouble.getCell(
                                    row,
                                    col,
                                    handleOutOfMemoryError );
                            if ( Double.isNaN( value ) ) {
                                pw.print( noDataValueBigDecimalString + " " );
                                System.out.println( "Warning!!! NaN encountered at row " + row + ", column " + col + " set to noDataValue " + noDataValueBigDecimalString + "." );
                            } else {
                                if ( value == noDataValue ) {
                                    pw.print( noDataValueBigDecimalString + " " );
                                } else {
                                    pw.print( value + " " );
                                }
                            }
                        } catch ( OutOfMemoryError oome0 ) {
                            grid2DSquareCellDouble.clearMemoryReserve();
                            grid2DSquareCellDouble.swapToFileGrid2DSquareCellChunks( handleOutOfMemoryError );
                            //pw.print( grid.getCell( row, col ) + " " );
                            value = grid2DSquareCellDouble.getCell(
                                    row,
                                    col,
                                    handleOutOfMemoryError );
                            if ( Double.isNaN( value ) ) {
                                pw.print( noDataValueBigDecimalString + " " );
                                System.out.println( "Warning!!! NaN encountered at row " + row + ", column " + col + " set to noDataValue " + noDataValueBigDecimalString + "." );
                            } else {
                                if ( value == noDataValue ) {
                                    pw.print( noDataValueBigDecimalString + " " );
                                } else {
                                    pw.print( value + " " );
                                }
                            }
                        }
                    }
                    pw.println( "" );
                }
            }
            // Close output
            pw.flush();
            pw.close();
        } catch ( FileNotFoundException fnfe0 ) {
            System.out.println( fnfe0.toString() );
            fnfe0.printStackTrace();
        }
        return file;
    }
}